;;; -*- Mode: Scheme -*-

;;;	Author: Sam Rushing <rushing@nightmare.com>
;;;	$Id: socket-object.scm,v 1.2 1997/03/10 07:41:40 rushing Exp $

;;; sockets, fd-sets, and select for gambit, using eotya/yasos

(include "eotya.scm")

;; probably should have a base class for pointer objects, so that 'free'
;; is a generic operation.  could also have some sort of 'access' operation
;; that would safely avoid NULL pointers... which would allow us to safely
;; 'free' malloc'd memory.

;; generic 'free' operation
(define-operation (free	object))

;; ======================================================================
;; socket object
;; ======================================================================

(define-predicate socket?)

;; attributes
(define-operation (get-fd	self		))
(define-operation (get-address 	self		))
;; operations
(define-operation (set-blocking	self bool	))
(define-operation (bind		self address	))
(define-operation (connect	self address	))
(define-operation (listen	self backup	))
(define-operation (accept	self		))
(define-operation (recv		self size	))
(define-operation (send		self string	))
(define-operation (close	self		))

;; implementation

(define (make-socket type family protocol . provided-socket)
  (let ((socket
	 (if (null? provided-socket)
	     (##socket-make type family protocol)
	     (car provided-socket))))
    (if socket
	(object
	 ;; predicate
	 ((socket?	self) #t)
	 ;; attributes
	 ((get-fd	self) (##socket-fd socket))
	 ((get-address	self) (##socket-address socket))
	 ;; operations
	 ((set-blocking	self block)
	  (##socket-set-blocking socket block))
	 ((bind self address)
	  (##socket-bind socket address))
	 ((connect self address)
	  (##socket-connect socket address))
	 ((listen self backlog)
	  (##socket-listen socket backlog))
	 ((accept self)
	  (let ((client (##socket-accept socket)))
	    (if client
		;; if we get back a low-level socket object,
		;; wrap it as an object before returning it
		(make-socket type family protocol client)
		client)))
	 ((recv self buffer-size)
	  (##socket-recv socket buffer-size))
	 ((send self string)
	  (##socket-send socket string (string-length string)))
	 ((close self)
	  (##socket-close socket)
	  (free self))
	 ((free self)
	  (##socket-free socket))
	 )
	#f
	)
    )
  )


;; ======================================================================
;; low-level file descriptor set
;; ======================================================================

(define-predicate fd-set?)
(define-operation (zero!	self))
(define-operation (set-fd!  	self fd))
(define-operation (is-fd-set?  	self fd))

;; maybe fd-sets should be be embedded in the selector object?
(define-operation (##get-fd-set	self))

(define (make-fd-set size)
  ;; WARNING: size param is ignored for now, and will probably only
  ;; be implemented for win32 (if ever).  You will need to redefine
  ;; FD_SETSIZE at the very top of scm_socket.h in order to get
  ;; something larger than the usual default.
  ;; For example, win32 defaults to 64, linux to 256.
  ;; [I think OSF/1 on the DEC Alpha may be as high as 4096!]

  (let ((fd-set (##fd-set-make))
	(count 0))
    (if fd-set
	(object
	 ((fd-set?	self) #t)
	 ((zero!	self)
	  (##fd-set-zero fd-set)
	  (set! count 0))
	 ((set-fd!	self fd)
	  (##fd-set fd-set fd)
	  (set! count (+ 1 count)))
	 ((is-fd-set?	self fd)
	  (##fd-isset fd-set fd))
	 ((free		self)
	  (##fd-set-free fd-set))
	 ((##get-fd-set	self)
	  fd-set)
	 )
	#f
	)))

(define-predicate selector?)

;; a timeout of #f means 'block until something fires'
(define-operation (select read-objects write-objects expt-objects timeout))

(define (make-selector size)
  (let ((read-fd-set  (make-fd-set size))
	(write-fd-set (make-fd-set size))
	(expt-fd-set  (make-fd-set size))
	)

    ;; given a list of selectable objects, modify an fd-set object to
    ;; represent it.  we also compute the largest file descriptor,
    ;; which we will pass to select.  [unfortunately, this parameter
    ;; is ignored on win32, so calculating it is a waste of time on
    ;; that platform].

    (define (list-set->fd-set list-set fd-set)
      (let ((max-fd 0))
	(zero! fd-set)
	(for-each
	 (lambda (object)
	   (let ((fd (get-fd object)))
	     (if (> fd max-fd)
		 (set! max-fd fd))
	     (set-fd! fd-set fd)))
	 list-set)
	max-fd))

    ;; given a list of selectable objects, and an fd-set
    ;; representing a subset of them, return a list of
    ;; selected objects.

    (define (fd-set->list-set list-set fd-set)
      (let ((result '()))
	(for-each
	 (lambda (object)
	   (if (is-fd-set? fd-set (get-fd object))
	       (set! result (cons object result))))
	 list-set)
	result))

    (define (display-fd-set set)
      (display "[")
      (for-each (lambda (x) (display " ") (display (get-fd x))) set)
      (display "]"))
    
    (define (display-fd-sets r w e)
      (display "r: ")  (display-fd-set r)
      (display " w: ") (display-fd-set w)
      (display " e: ")  (display-fd-set e)
      (newline))
    
    (if (and read-fd-set write-fd-set expt-fd-set)
	(object
	 ((selector? self) #t)
	 ((select self read-objects write-objects expt-objects timeout)

	  ;(display "before ")
	  ;(display-fd-sets read-objects write-objects expt-objects)

	  (let ((forever? (not (number? timeout)))
		(timeout (if (eq? timeout #f) 0.0 timeout))
		(max-fd (max
			 (list-set->fd-set read-objects read-fd-set)
			 (list-set->fd-set write-objects write-fd-set)
			 (list-set->fd-set expt-objects expt-fd-set))))
	    (##select
	     (+ max-fd 1)
	     (##get-fd-set read-fd-set)
	     (##get-fd-set write-fd-set)
	     (##get-fd-set expt-fd-set)
	     forever?
	     timeout)
	    (let ((result (list (fd-set->list-set read-objects read-fd-set)
				(fd-set->list-set write-objects write-fd-set)
				(fd-set->list-set expt-objects expt-fd-set))))
	      ;(display "after ")
	      ;(apply display-fd-sets result)
	      result)
	    ))
	 )
	#f)))
